/*************************************************************
UCF EEL4915 SPRING 2018, Team-8 Laser Skeet project

This is a modified version of Texas instruments project file to add our 
Functionality.  TI licensing terms apply to any SW use.

**************************************************************/
/******************************************************************************

 @file board_key.c

 @brief This file contains the interface to the Launchpad Key Service

 <BR>
 NOTE:  The launchpad only has 2 buttons, so only KEY_LEFT and KEY_RIGHT are
        enabled.
 <BR>
 Group: WCS LPC
 Target Device: CC13xx

 ******************************************************************************
 
 Copyright (c) 2016-2017, Texas Instruments Incorporated
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 *  Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

 *  Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

 *  Neither the name of Texas Instruments Incorporated nor the names of
    its contributors may be used to endorse or promote products derived
    from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 ******************************************************************************
 Release Name: simplelink_cc13x0_sdk_1_50_00_
 Release Date: 2017-09-27 23:49:29
 *****************************************************************************/

/******************************************************************************
 Includes
 *****************************************************************************/

#include <stdbool.h>

#include "timer.h"
#include "board_key.h"

/******************************************************************************
 Local Variables
 *****************************************************************************/

/* Value of keys Pressed */
static uint8_t keysPressed;

/* Key debounce clock */
static Clock_Struct keyChangeClock;

// UCF team 8 project addition for timed action delay for simulated flight time
static Clock_Struct sendDelayClock;

/* Pointer to application callback */
static Board_Key_keysPressedCB_t appKeyChangeHandler = NULL;

/*
 Create the MSA KEY pin table. This will override the key attributes in
 BoardGpioInitTable[].
 */
static PIN_Config keyPinTable[] = {
      Board_BUTTON0 | PIN_INPUT_EN | PIN_PULLUP | PIN_IRQ_NEGEDGE,
      Board_BUTTON1 | PIN_INPUT_EN | PIN_PULLUP | PIN_IRQ_NEGEDGE,
      Board_DIO15 | PIN_INPUT_EN | PIN_PULLUP | PIN_IRQ_NEGEDGE,
      PIN_TERMINATE /* Terminate list */
    };

/* KEY pin state */
static PIN_State keyPinState;

/* KEY Pin Handle */
static PIN_Handle keyPinHandle;

/******************************************************************************
 Local Functions
 *****************************************************************************/
static void board_key_changeHandler(UArg a0);
static void board_key_keyFxn(PIN_Handle keyPinHandle, PIN_Id keyPinId);
static uint8_t board_key_getValues(void);

// UCF Team 8 addition
static void board_key_delayedSendHandler(UArg a0);

/******************************************************************************
 Public Functions
 *****************************************************************************/

/*!
 Enable interrupts for keys on PIN.

 Public function defined in board_key.h
 */
uint8_t Board_Key_initialize(Board_Key_keysPressedCB_t appKeyCB)
{
    uint8_t keyState;

    /* Setup KEY ISR */
    keyPinHandle = PIN_open(&keyPinState, keyPinTable);

    /* Get current key state */
    keyState = board_key_getValues();

    /* Register Callbacks */
    PIN_registerIntCb(keyPinHandle, board_key_keyFxn);

    /* Setup keycallback for keys */
    Timer_construct(&keyChangeClock, board_key_changeHandler,
                    (KEY_DEBOUNCE_TIMEOUT),
                    0, false, 0);

    // UCF Team 8 addition - laser delay timer
    Timer_construct(&sendDelayClock, board_key_delayedSendHandler,
                    (KEY_DELAY_TIMEOUT),
                    0, false, 0);

    /* Set the application callback */
    appKeyChangeHandler = appKeyCB;

    return(keyState);
}

/*!
 * @brief       Interrupt handler for a Key press
 *
 * @param       keyPinHandle - PIN Handle
 * @param       keyPinId - PIN ID
 */
static void board_key_keyFxn(PIN_Handle keyPinHandle, PIN_Id keyPinId)
{
    (void)keyPinHandle;
/*
    if(keyPinId == Board_BUTTON0)
    {
        keysPressed |= KEY_LEFT;
    }
    else if(keyPinId == Board_BUTTON1)
    {
        keysPressed |= KEY_RIGHT;


    }
    else if(keyPinId == Board_DIO15)
    {
        keysPressed |= KEY_RELOAD;
    }

    if(Timer_isActive(&keyChangeClock) != true)
    {
        Timer_start(&keyChangeClock);
    }
*/

    // if laser control - delayed send for simulated shot flight time
    // if BUTTON0 and not set to send
     if(  (keyPinId == Board_BUTTON0) &&
         !(keysPressed & KEY_LEFT) )
     {
        keysPressed |= KEY_LEFT;
        Timer_start(&sendDelayClock);
     }
     // if BUTTON1 and not set to send
     else if( (keyPinId == Board_BUTTON1) &&
             !(keysPressed & KEY_RIGHT) )
     {
        keysPressed |= KEY_RIGHT;
        Timer_start(&keyChangeClock);
     }
     // if BUTTON3 and not set to send
     else if( (keyPinId == Board_DIO15) &&
         !(keysPressed & KEY_RELOAD) )
     {
        keysPressed |= KEY_RELOAD;
        Timer_start(&keyChangeClock);
     }


}

/*!
 * UCF Team 8 addition
 * @brief       Handler for Delayed action for Laser flight time simulation
 *
 * @param       UArg a0 - ignored
 *
 * send key notification after delay
 */
static void board_key_delayedSendHandler(UArg a0)
{
    if(PIN_getInputValue(Board_BUTTON0) == false)
    {
    if(appKeyChangeHandler != NULL)
    {
        /* Notify the application */
        (*appKeyChangeHandler)(KEY_LEFT);
    }

    // start debounce timer
    Timer_start(&keyChangeClock);
    }
    else
    {
        keysPressed = (keysPressed & ~KEY_LEFT);
    }
}


/*!
 * @brief       Handler for key change
 *
 * @param       UArg a0 - ignored
 *
 * executes on debounce timeout
 */
static void board_key_changeHandler(UArg a0)
{
    uint8_t keys;
    keys = board_key_getValues();

    // clear button0 it is sent early with special timing,
    keysPressed = (keysPressed & ~KEY_LEFT);
    keys = (keys & ~KEY_LEFT);

    // confirm button to avoid double hit on button release
    if(keys == keysPressed )
    {
        /* Notify the application */
        if((appKeyChangeHandler != NULL) && keys)
        {
            (*appKeyChangeHandler)(keysPressed);
        }
    }

    /* Clear keys to allow new key press (Debounce time expired)*/
    keysPressed = 0;

}

/*!
 * @brief       Get the current value for all the keys
 *
 * @return      bit mapped representation of all keys
 */
static uint8_t board_key_getValues(void)
{
    uint8_t keys = 0;

    if(PIN_getInputValue(Board_BUTTON0) == false)
    {
        keys |= KEY_LEFT;
    }

    if(PIN_getInputValue(Board_BUTTON1) == false)
    {
        keys |= KEY_RIGHT;
    }

    if(PIN_getInputValue(Board_DIO15) == false)
    {
        keys |= KEY_RELOAD;
    }

    return(keys);
}

